unit rkSmartPath;

interface

uses
  Windows, SysUtils, Classes, Controls, ExtCtrls, Graphics, Forms, Messages,
  ShellAPI, ImgList, Dialogs, Menus, StdCtrls;

const
  CM_REFRESH = WM_USER + 2000; // Custom Message...

type
  TspItemState = (spNormal, spButtonPressed, spArrowPressed, spButtonHot, spArrowHot);
  TspPaths = set of (spDesktop, spDocuments, spPictures, spMusic, spVideos);

  TrkSmartPath = class( TCustomControl )
  private
    popMenu: TPopupMenu;
    barImageList: TImageList;
    strMenu: TStringList;
    FCustomPopup: TPopupMenu;
    FNewFolder: String;
    FEditPath: TEdit;
    FAuto: Boolean;
    FBorderStyle: TBorderStyle;
    FColorStart: TColor;
    FColorEnd: TColor;
    FColorBorder: TColor;
    FBrightness: Integer;
    FContrast: Integer;
    FCenter: Boolean;
    FReverse: Boolean;
    FMainBmp: TBitmap;
    FFont: TFont;
    FHotPt: TPoint;
    FPath: String;
    FPaths: TStringList;
    FCustomPaths: TStringList;
    FPopUp: Boolean;
    FState: TspItemState;
    FPathIdx: Integer;
    FPathsAdd: TspPaths;
    FPopPath: String;
    FMClick: Boolean;
    FOnPathChanged: TNotifyEvent;
    procedure WMERASEBKGND(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMIsToolControl(var Message: TMessage); message CM_ISTOOLCONTROL;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetFont(Value: TFont);
    procedure PaintPath;
    procedure SetPath(Value: String);
    procedure SetCustomPaths(Value: TStringList);
    procedure SetPathsToAdd(Value: TspPaths);
    function  CalcPath: Integer;
    procedure CMRefresh(var Message: TMessage); message CM_REFRESH;
    procedure CalcState;
    procedure EditExit(Sender: TObject);
    procedure EditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure EditKeyPress(Sender: TObject; var Key: Char);

    { Private declarations }
  protected
    { Protected declarations }
    FPopPos: TPoint;
    FPathOff: Integer;
    FPathGlyph: Integer;
    FDoCreate: Boolean;
    // auto handling routines
    procedure popMenuPopup(Sender: TObject);
    procedure popMenuClick(Sender: TObject);
    // end auto handling routines

    procedure PaintWindow(DC: HDC); override;
    procedure Resize; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  public
    { Public declarations }
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    function GetSpecialFolderPath(folder : integer) : string;
    function SysImageListHandle(const Path: string;  const WantLargeIcons: Boolean): Windows.THandle;
    procedure DoEdit(CreateFolder: Boolean);
    property PopPath: String Read FPopPath;
    property PathOff: Integer Read FPathOff;
    property PathParts: TStringList Read FPaths;
  published
    { Published declarations }
    property NewFolderName: String Read FNewFolder Write FNewFolder;
    property OnPathChanged: TNotifyEvent Read FOnPathChanged Write FOnPathChanged;
    property Path: String Read FPath Write SetPath;
    property SpecialFolders: TspPaths read FPathsAdd write SetPathsToAdd;
    property Favourites: TStringList Read FCustomPaths Write SetCustomPaths;
    property PathGlyph: Integer Read FPathGlyph default -1;
    property Auto: Boolean Read FAuto Write FAuto default True;
    property CustomPopupMenu: TPopupMenu Read FCustomPopup Write FCustomPopup;
    property Align;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelOuter;
    property BevelKind;
    property BevelWidth;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsNone;
    property Font: TFont read FFont write SetFont;
    property BorderWidth;
    property Canvas;
    property Left;
    property Top;
    property Width;
    property Height;
    property ParentBackground;
    property PopupMenu;
    property ShowHint;
    property OnEnter;
    property OnExit;
    property OnResize;
    property OnClick;
    property OnDblClick;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnKeyDown;
    property OnKeyUp;
  end;

procedure Register;

implementation

uses
  ShlObj, SHFolder;

// helper routines for auto

function TrkSmartPath.GetSpecialFolderPath(folder : integer) : string;
const
  SHGFP_TYPE_CURRENT = 0;
var
  path: array [0..MAX_PATH] of char;
begin
  if SUCCEEDED(SHGetFolderPath(0,folder,0,SHGFP_TYPE_CURRENT,@path[0])) then
    Result := path
  else
    Result := '';
end;

// Gets handle to system image list for a specified path. WantLargeIcons
// determines if image list contains large or small icons.
function TrkSmartPath.SysImageListHandle(const Path: string; const WantLargeIcons: Boolean): Windows.THandle;
var
  FI: ShellAPI.TSHFileInfo;
  Flags: Windows.UINT;
begin
  Flags := ShellAPI.SHGFI_SYSICONINDEX or ShellAPI.SHGFI_ICON;
  if WantLargeIcons then
    Flags := Flags or ShellAPI.SHGFI_LARGEICON
  else
    Flags := Flags or ShellAPI.SHGFI_SMALLICON;
  Result := ShellAPI.SHGetFileInfo(PChar(Path), 0, FI, SizeOf(FI), Flags);
end;

function AddBackSlash( const AString: string ): string;
begin
  if ( AString = '' ) then Result := ''
  else begin
    if (AString[Length( AString )] <> '\') then
      Result := AString + '\'
    else Result := AString;
  end;
end;

// returns a unique name for a new folder with a number if necessary parameter
// Path must end with a backslash
function GetNewFolderName(const Path, FolderName: string ): string;
var
  i: integer;
begin
  Result := Path + FolderName;
  i := 0;
  while DirectoryExists(result) do begin
    inc(i);
    Result := Path + FolderName + ' ' + IntToStr(i);
  end;
end;

procedure TrkSmartPath.DoEdit(CreateFolder: Boolean);
var
  i: Integer;
  txt: String;
begin
  FDoCreate:= CreateFolder;
  FEditPath.Font.Assign(FFont);
  FEditPath.Clear;
  FEditPath.Width:= ClientWidth;
  FEditPath.AutoSelect:= not CreateFolder;
  if CreateFolder then begin
    txt:= Path;
    i:= Length(Path) + 1;
    txt:= txt + '\';
    txt:= GetNewFolderName(txt, FNewFolder);
    FEditPath.Text:= txt;
    FEditPath.SelStart:= i;
    FEDitPath.SelLength:= Length(txt) - i;
  end else
    FEditPath.Text:= Path;
  FEditPath.Visible:= True;
  FEDitPath.SetFocus;
end;

procedure TrkSmartPath.EditKeyDown(Sender: TObject; var Key: Word; Shift:
  TShiftState);
var
  done: Boolean;
begin
  if key = VK_Escape then done:= True
  else if key = VK_Return then begin
    done:= True;
    if FDoCreate then begin
      if not CreateDir(FEditPath.Text) then
        ShowMessage('Sorry, I could not create that directory. ErrorCode: ' +
          IntToStr(GetLastError));
    end;
    Path:= FEditPath.Text;
  end else done:= False;
  FEditPath.Visible := not done;
  //IsEditing := EditInline.Visible;
  if done then SetFocus;
end;

procedure TrkSmartPath.EditKeyPress(Sender: TObject; Var Key: Char);
begin
  if (Key = #13) or (Key = #27) then Key:= #0;
end;

procedure TrkSmartPath.EditExit(Sender: TObject);
begin
  FEditPath.Visible := False;
  //IsEditing:= False;
  SetFocus;
  Invalidate;
end;

procedure MultiSzToStrings(const MultiSz: PChar; const Strings: Classes.TStrings);
var
  P: PChar;
begin
  if not Assigned(MultiSz) then Exit;
  P := MultiSz;
  while P^ <> #0 do begin
    Strings.Add(P);
    Inc(P, SysUtils.StrLen(P) + 1);
  end;
end;

procedure TrkSmartPath.popMenuPopup(Sender: TObject);
const
  Flag : Cardinal = SHGFI_SMALLICON or SHGFI_LARGEICON or SHGFI_SYSICONINDEX or
    SHGFI_TYPENAME or SHGFI_DISPLAYNAME;
  SSI: Integer = SizeOf(TSHFileInfo);
var
  AMenuItem: TMenuItem;
  SFI: TSHFileInfo;
  SR: TSearchRec;
  IsFound: Boolean;
  i, j, idx: integer;
  FName, s: string;
  Drives: PChar;
  BufSize: integer;
  List: TStringList;

  procedure AddMenuItem(PName, FName: String; Icon: Integer);
  begin
    AMenuItem := TMenuItem.Create(popMenu);
    AMenuItem.Caption:= FName;
    AMenuItem.OnClick := popMenuClick;
    AMenuItem.ImageIndex:= Icon;
    AmenuItem.Tag:= Idx;
    Inc(idx);
    StrMenu.Add(PName);
    popMenu.Items.Add(AMenuItem);
  end;

begin
  Screen.Cursor := crHourglass;
  for i := popMenu.Items.Count - 1 downto 0 do popMenu.Items[i].Free;
  StrMenu.Clear;
  idx:= 0;
  if FPopPath = 'HomeFolder' then begin
    if FPathOff > 0 then begin
      for i:= FPathOff - 1 downto 0 do begin
        FName:= '';
        for j:= 0 to i do FName:= FName + FPaths[j] + '\';
        SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
        AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
      end;
      AddMenuItem('- ', '-', -1);
    end;

    AddMenuItem('*', 'New Folder', -1);
    AddMenuItem('- ', '-', -1);

    if spDesktop in FPathsAdd then begin
      FName:= GetSpecialFolderPath(CSIDL_DESKTOPDIRECTORY);
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spDocuments in FPathsAdd then begin
      FName:= GetSpecialFolderPath(CSIDL_PERSONAL);
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spPictures in FPathsAdd then begin
      FName:= GetSpecialFolderPath(CSIDL_MYPICTURES);
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spMusic in FPathsAdd then begin
      FName:= GetSpecialFolderPath(CSIDL_MYMUSIC);
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    if spVideos in FPathsAdd then begin
      FName:= GetSpecialFolderPath(CSIDL_MYVIDEO);
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;

    AddMenuItem('- ', '-', -1);

    // Add favourites
    for i:= 0 to FCustomPaths.Count - 1 do begin
      FName:= FCustomPaths[i];
      s:= ExtractFileName(Fname);
      SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;

    AddMenuItem('- ', '-', -1);

    List:= TStringList.Create;
    BufSize := Windows.GetLogicalDriveStrings( 0, nil );
    GetMem( Drives, BufSize * SizeOf( Char ) ); // Delphi 2007 or 2009
    try
      if Windows.GetLogicalDriveStrings( BufSize, Drives ) = 0 then SysUtils.RaiseLastOSError;
      MultiSzToStrings( Drives, List );
    finally
      FreeMem( Drives );
    end;
    for i:= 0 to List.Count -1 do begin
      FName := List[i];
      SHGetFileInfo( PChar( FName ), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
      AddMenuItem(FName, SFI.szDisplayName, SFI.iIcon);
    end;
    List.Free;
  end else begin
    IsFound := FindFirst(FPopPath + '*.*', faAnyFile - faHidden, SR) = 0;
    while IsFound do
    begin
      if (SR.Name <> '.') and (SR.Name <> '..') then
      begin
        if ((SR.Attr and faDirectory) <> 0) then
        begin
          FName := FPopPath + SR.Name;
          SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI, SSI, Flag);
          AddMenuItem(SR.Name, SFI.szDisplayName, SFI.iIcon);
        end;
      end;
      IsFound := FindNext(SR) = 0;
    end;
    FindClose( SR );
  end;
  Screen.Cursor := crDefault;
end;

procedure TrkSmartPath.popMenuClick(Sender: TObject);
var
  txt: String;
begin
  if FPopPath = 'HomeFolder' then begin
    txt:= StrMenu[(Sender as TMenuItem).Tag];
    if txt = '*' then DoEdit(True) else Path:= txt
  end else
    Path:= FPopPath + StrMenu[(Sender as TMenuItem).Tag];
end;
// end auto helper routines

procedure TrkSmartPath.PaintWindow(DC: HDC);
begin
  Canvas.Lock;
  try
    Canvas.Handle := DC;
    try
      FMainBmp.Canvas.Font:= FFont;
      PaintPath;
    finally
      Canvas.Handle := 0;
    end;
  finally
    Canvas.Unlock;
  end;
end;

procedure FillGradient(const Canvas: TCanvas; const ARect: TRect;
  const StartColor, EndColor: TColor;  const Vertical: Boolean);
type
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..1024] of TRGBTriple;
  TGradientColors = array[0..255] of TRGBTriple;
var
  rc1, gc1, bc1, rc2, gc2, bc2, rc3, gc3, bc3,
    y1, i, GSize: Integer;
  Row: PRGBTripleArray;
  GradCol: TRGBTriple;
  GradientBmp: TBitmap;
begin
  rc2 := GetRValue(ColorToRGB(StartColor));
  gc2 := GetGValue(ColorToRGB(StartColor));
  bc2 := GetBValue(ColorToRGB(StartColor));
  rc1 := GetRValue(ColorToRGB(EndColor));
  gc1 := GetGValue(ColorToRGB(EndColor));
  bc1 := GetBValue(ColorToRGB(EndColor));
  rc3 := rc1 + (((rc2 - rc1) * 15) div 9);
  gc3 := gc1 + (((gc2 - gc1) * 15) div 9);
  bc3 := bc1 + (((bc2 - bc1) * 15) div 9);
  if rc3 < 0 then rc3 := 0 else if rc3 > 255 then rc3 := 255;
  if gc3 < 0 then gc3 := 0 else if gc3 > 255 then gc3 := 255;
  if bc3 < 0 then bc3 := 0 else if bc3 > 255 then bc3 := 255;
  GradientBMP := TBitmap.Create;
  GradientBmp.PixelFormat := pf24bit;
  if Vertical then
  begin
    GradientBmp.Width := 1;
    GradientBmp.Height := (ARect.Bottom - ARect.Top) - 1;
    GSize := GradientBmp.Height;
    y1 := GSize div 2;
    for i := 0 to y1 - 1 do
    begin
      Row := PRGBTripleArray(GradientBmp.ScanLine[i]);
      GradCol.rgbtRed := Byte(rc1 + (((rc2 - rc1) * (i)) div y1));
      GradCol.rgbtGreen := Byte(gc1 + (((gc2 - gc1) * (i)) div y1));
      GradCol.rgbtBlue := Byte(bc1 + (((bc2 - bc1) * (i)) div y1));
      Row[0] := GradCol;
    end;
    if rc2 > rc1 then
    begin
      rc3 := rc2;
      gc3 := gc2;
      bc3 := bc2;
    end;
    for i := y1 to GSize - 1 do
    begin
      Row := PRGBTripleArray(GradientBmp.ScanLine[i]);
      GradCol.rgbtRed := Byte(rc3 + (((rc2 - rc3) * (i)) div GSize));
      GradCol.rgbtGreen := Byte(gc3 + (((gc2 - gc3) * (i)) div GSize));
      GradCol.rgbtBlue := Byte(bc3 + (((bc2 - bc3) * (i)) div GSize));
      Row[0] := GradCol;
    end;
    Canvas.StretchDraw(ARect, GradientBmp);
  end else
  begin
    GradientBmp.Width := (ARect.Right - ARect.Left) - 1;
    GradientBmp.Height := 1;
    GSize := GradientBmp.Width;
    y1 := GSize div 2;
    Row := PRGBTripleArray(GradientBmp.ScanLine[0]);
    for i := 0 to y1 - 1 do
    begin
      GradCol.rgbtRed := Byte(rc1 + (((rc2 - rc1) * (i)) div y1));
      GradCol.rgbtGreen := Byte(gc1 + (((gc2 - gc1) * (i)) div y1));
      GradCol.rgbtBlue := Byte(bc1 + (((bc2 - bc1) * (i)) div y1));
      Row[i] := GradCol;
    end;
    if rc2 > rc1 then
    begin
      rc3 := rc2;
      gc3 := gc2;
      bc3 := bc2;
    end;
    for i := y1 to GSize - 1 do
    begin
      GradCol.rgbtRed := Byte(rc2 + (((rc3 - rc2) * (i)) div GSize));
      GradCol.rgbtGreen := Byte(gc2 + (((gc3 - gc2) * (i)) div GSize));
      GradCol.rgbtBlue := Byte(bc2 + (((bc3 - bc2) * (i)) div GSize));
      Row[i] := GradCol;
    end;
    Canvas.StretchDraw(ARect, GradientBmp);
  end;
  GradientBmp.Free;
end;

procedure TrkSmartPath.CalcState;
var
  w: Integer;
  txt: String;
  Arrow, Done: Boolean;
  PathPos: Integer;
  xp, cw: Integer;
  popX, popY: Integer;
  s: String;
begin
  FState:= spNormal;
  xp:= 22;
  Arrow:= True;
  Done:= False;
  s:= '';
  FPathIdx:= -1;
  PathPos:= FPathOff - 1;
  for w:= 0 to PathPos do s:= s + FPaths[w] + '\';
  while (not Done) do begin
    if Arrow then begin
      if (FHotPt.X > xp) and (FHotPt.X < xp + 14) then begin
        if FMClick then FState:= spArrowPressed else FState:= spArrowHot;
        popx:= xp + 14 - 22;
        popy:= ClientHeight - 2;
        FPopPos:= ClientToScreen(Point(popX, popY));
        FPopPath:= 'HomeFolder';
        FPathIdx:= 0;
      end;
      xp:= xp + 14;
      Arrow:= False;
    end else begin
      inc(PathPos);
      txt:= FPaths[PathPos];
      s:= s + txt + '\';
      w:= FMainBmp.Canvas.TextWidth(txt);
      cw:= w + 9 + 14;
      if (FHotPt.X > xp) and (FHotPt.X < (xp + cw)) then begin
        if not FMClick then begin
          if (FHotPt.X > xp) and (FHotPt.X < (xp + cw) - 14) then
            FState:= spButtonHot
          else
            FState:= spArrowHot;
        end;
        if FMClick then begin
          FPathIdx:= PathPos;
          if (FHotPt.X > xp) and (FHotPt.X < (xp + cw) - 14) then
            FState:= spButtonPressed
          else
            FState:= spArrowPressed;
        end;
        popx:= xp + cw - 22;
        popy:= ClientHeight - 2;
        FPopPos:= ClientToScreen(Point(popX, popY));
        FPopPath:= s;
      end;
      xp:= xp  + cw;
    end;
    Done:= (xp > ClientWidth) or (PathPos = FPaths.Count - 1);
  end;
end;

procedure TrkSmartPath.PaintPath;
var
  w, ax, ay: Integer;
  txt: String;
  R, T: TRect;
  Arrow, Done, bool, HA: Boolean;
  PathPos, TextOff: Integer;
  xp, cw: Integer;
  PathCanvas: TCanvas;

  procedure PaintArrow(Canvas: TCanvas; x, y: Integer; Down: Boolean);
  begin
    with Canvas do begin
      Pen.Color:= clBlack;
      if Down then begin
        MoveTo(x + 3, y + 2); LineTo(x + 7, y + 2); LineTo(x + 5, Y + 4);
        LineTo(x + 3, y + 2); LineTo(x + 6, y + 4); MoveTo(x + 2, y + 1);
        LineTo(x + 9, y + 1);
      end else begin
        MoveTo(x + 4, y); LineTo(x + 4, y + 4); LineTo(x + 6, y + 2);
        LineTo(x + 4, y); LineTo(x + 5, y + 3); LineTo(x + 3, y - 1);
        LineTo(x + 3, y + 6);
      end;
    end;
  end;

begin
  FMainBmp.Width := ClientWidth;
  FMainBmp.Height := ClientHeight;
  PathCanvas:= FMainBmp.Canvas;
  with PathCanvas do begin
    Brush.Style:= bsClear;
    Pen.Color:= clWindow;
    R:= ClientRect;
    Rectangle(R);
    Brush.Color:= clWindow;
    Brush.Style:= bsSolid;
    Pen.Color:= clSilver;
    R:= ClientRect;
    InflateRect(R, -1, -1);
    Rectangle(R);
  end;
  if (FPathGlyph <> -1) then barImageList.Draw(PathCanvas, 4, 3, FPathGlyph);
  InflateRect(R, -1, -1);
  ay:= ((ClientHeight - 7) div 2) + 1;
  xp:= 22;
  Arrow:= True;
  Done:= False;
  HA:= False;
  PathPos:= FPathOff - 1;
  while (not Done) do begin
    TextOff:= 0;
    T.Top:= R.Top + 1;
    T.Bottom:= R.Bottom - 1;
    if Arrow then begin
      ax:= xp + 3;
      if (FHotPt.X > xp) and (FHotPt.X < xp + 14) then begin
        T.Left:= xp;
        T.Right:= T.Left + 15;
        PathCanvas.Pen.Color:= $B17F3C;
        PathCanvas.MoveTo(T.Left, R.Top);
        PathCanvas.LineTo(T.Left, R.Bottom);
        PathCanvas.MoveTo(T.Right - 1, R.Top);
        PathCanvas.LineTo(T.Right - 1, R.Bottom);
        T.Left:= T.Left + 2;
        T.Right:= T.Right - 2;
        if not FMClick then
           FillGradient(PathCanvas, T, $F4D9A6, $FDF6EA, True)
        else begin
          InflateRect(T, 1, 1);
          FillGradient(PathCanvas, T, $FADFAC, $FDF6EA, True);
          InflateRect(T, -1, -1);
          PathCanvas.Pen.Color:= $927764;
          PathCanvas.MoveTo(T.Left - 1 , R.Bottom - 1);
          PathCanvas.LineTo(T.Left - 1, R.Top);
          PathCanvas.LineTo(T.Right + 1, R.Top);
          PathCanvas.MoveTo(T.Right, R.Top + 1);
          PathCanvas.Pen.Color:= $EACF9C;
          PathCanvas.LineTo(T.Left, R.Top + 1);
          PathCanvas.LineTo(T.Left, R.Bottom - 1);
          PathCanvas.LineTo(T.Right + 1, R.Bottom - 1);
        end;
      end;
      PaintArrow(PathCanvas, ax, ay, FPathIdx = 0);
      Arrow:= False;
      xp:= xp + 14;
    end else begin
      inc(PathPos);
      txt:= FPaths[PathPos];
      w:= PathCanvas.TextWidth(txt);
      cw:= w + 9 + 14;
      if (FHotPt.X > xp) and (FHotPt.X < (xp + cw)) then begin
        T.Left:= xp;
        T.Right:= T.Left + w + 9;
        bool:= (FHotPt.X > xp) and (FHotPt.X < (xp + cw) - 14);
        if (bool) or (FMClick) then PathCanvas.Pen.Color:= $B17F3C
        else PathCanvas.Pen.Color:= $8F8F8F;
        if FMClick then begin
          TextOff:= 1;
          T.Top:= R.Top;
          T.Bottom:= R.Bottom;
          T.Left:= T.Left + 1;
          T.Right:= T.Right + 14;
          FillGradient(PathCanvas, T, $FADFAC, $FDF6EA, True);
          PathCanvas.Pen.Color:= $927764;
          PathCanvas.MoveTo(T.Left + 1, R.Bottom - 1);
          PathCanvas.LineTo(T.Left + 1, R.Top);
          PathCanvas.LineTo(T.Right - 1, R.Top);
          PathCanvas.MoveTo(T.Right - 2, R.Top + 1);
          PathCanvas.Pen.Color:= $EACF9C;
          PathCanvas.LineTo(T.Left + 2, R.Top + 1);
          PathCanvas.LineTo(T.Left + 2, R.Bottom - 1);
          PathCanvas.MoveTo(T.Right - 12, R.Bottom - 1);
          PathCanvas.LineTo(T.Right - 12, R.Top + 1);
          PathCanvas.Pen.Color:= $927764;
          PathCanvas.MoveTo(T.Right - 13, R.Bottom - 1);
          PathCanvas.LineTo(T.Right - 13, R.Top);
          PathCanvas.Pen.Color:= $B17F3C;
          PathCanvas.MoveTo(T.Left, R.Top);
          PathCanvas.LineTo(T.Left, R.Bottom);
          PathCanvas.MoveTo(T.Right - 1, R.Top);
          PathCanvas.LineTo(T.Right - 1, R.Bottom);
          PathCanvas.MoveTo(T.Right - 14, R.Top);
          PathCanvas.LineTo(T.Right - 14, R.Bottom);
          Ha:= True;
        end else begin
          PathCanvas.MoveTo(T.Left, R.Top);
          PathCanvas.LineTo(T.Left, R.Bottom);
          PathCanvas.Pen.Color:= $B17F3C;
          PathCanvas.MoveTo(T.Right - 1, R.Top);
          PathCanvas.LineTo(T.Right - 1, R.Bottom);
          T.Left:= T.Left + 2;
          T.Right:= T.Right - 2;
          if (bool) then
            FillGradient(PathCanvas, T, $FADFAC, $FDF6EA, True)
          else
            FillGradient(PathCanvas, T, $E3E3E3, $F2F2F2, True);
          T.Left:= xp + w + 8;
          T.Right:= T.Left + 15;
          PathCanvas.Pen.Color:= $B17F3C;
          PathCanvas.MoveTo(T.Left, R.Top);
          PathCanvas.LineTo(T.Left, R.Bottom);
          PathCanvas.MoveTo(T.Right - 1, R.Top);
          PathCanvas.LineTo(T.Right - 1, R.Bottom);
          T.Left:= T.Left + 2;
          T.Right:= T.Right - 2;
          FillGradient(PathCanvas, T, $F4D9A6, $FDF6EA, True);
        end;
      end;
      if ((xp + cw) < ClientWidth - 5) then begin
        T:= R;
        T.Left:= xp + 1 + TextOff;
        T.Right:= T.Left + w + 7;
        T.Top := T.Top + TextOff;
        T.Bottom := T.Bottom + TextOff;
        PathCanvas.Brush.Style:= bsClear;
        DrawText(PathCanvas.Handle, PChar(txt), Length(txt), T,
          DT_CENTER or DT_VCENTER or DT_SINGLELINE );
        ax:= T.Right + 3 - TextOff;
        T.Top := R.Top;
        PaintArrow(PathCanvas, ax, ay, HA);
        HA:= False;
      end;
      xp:= xp  + cw;
    end;
    Done:= (xp > ClientWidth) or (PathPos = FPaths.Count - 1);
  end;
  BitBlt(Canvas.Handle, 0, 0, FMainBmp.Width, FMainBmp.Height, FMainBmp.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TrkSmartPath.WMERASEBKGND(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TrkSmartPath.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrkSmartPath.CMCtl3DChanged(var Message: TMessage);
begin
  if NewStyleControls and (FBorderStyle = bsSingle) then RecreateWnd;
  inherited;
end;

procedure TrkSmartPath.CMIsToolControl(var Message: TMessage);
begin
  Message.Result := 1;
end;

procedure TrkSmartPath.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  Invalidate;
  inherited;
  if not (csLoading in ComponentState) then Resize;
end;

procedure TrkSmartPath.CMRefresh(var Message: TMessage);
begin
  if (csDestroying in ComponentState) then Exit;
  FPathIdx:= -1;
  FPopPath:= '';
  FPopUp:= False;
  FMClick:= False;
  FHotPt:= Point(-1, -1);
  FState:= spNormal;
  Invalidate;
  inherited;
end;

procedure TrkSmartPath.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TrkSmartPath.SetFont(Value: TFont);
begin
  if (FFont <> Value) then FFont.Assign(Value);
  Invalidate;
end;

procedure TrkSmartPath.SetPath(Value: String);
var
  SFI: TSHFileInfo;
begin
  if (FPath = Value) or not(DirectoryExists(Value)) then Exit;
  FPath:= Value;
  FPaths.Clear;
  FPaths.Text:= Trim(StringReplace(FPath, '\', #13, [rfReplaceAll]));
  ShGetFileInfo(PChar(FPath), 0, SFI, SizeOf(TSHFileInfo), SHGFI_SYSICONINDEX
    or SHGFI_ICON or SHGFI_DISPLAYNAME or SHGFI_TYPENAME or SHGFI_SMALLICON);
  FPathGlyph:=  SFI.iIcon;
  FMClick:= False;

  if not (csDesigning in ComponentState) then FPathOff:= CalcPath;

  Invalidate;
  if Assigned(FOnPathChanged) then FOnPathChanged(Self);
end;

procedure TrkSmartPath.SetCustomPaths(Value: TStringList);
begin
  FCustomPaths.Assign(Value);
end;

procedure TrkSmartPath.SetPathsToAdd(Value: TspPaths);
begin
  FPathsAdd:= Value;
end;

function TrkSmartPath.CalcPath : Integer;
var
  i, w: Integer;
begin
  Result:= 0;
  w:= 22 + 14; // Glyph + Arrow
  for i := 0 to FPaths.Count - 1 do
    w:= w + FMainBmp.Canvas.TextWidth(FPaths[i]) + 23; // Space and arrow
  // w now contains the width in pixels needed to show complete path
  // Now we check to see if it fits ...
  if w < ClientWidth - 8 then Exit; // if it does we are done :)
  // ok it did not fit so we must reduce it. Lets do it
  // How much are we talking about
  w:= w - (ClientWidth - 8);
  i:= 0;
  while (w > 0) and (i < FPaths.Count)do begin
    w:= w - (FMainBmp.Canvas.TextWidth(FPaths[i]) + 23);
    i:= i + 1;
  end;
  if w > 0 then Result:= FPaths.Count else Result:= i;
end;

constructor TrkSmartPath.Create( AOwner: TComponent );
var
  ADesktopFolder: string;
  TmpHandle: THandle;
begin
  inherited;
  ControlStyle := ControlStyle + [csOpaque, csAcceptsControls, csReplicatable, csNeedsBorderPaint];
  Width:= 150;
  Height:= 25;
  FState:= spNormal;
  FMClick:= False;
  FPopPath:= '';
  FPath:= '';
  FPaths:= TStringList.Create;
  FPathOff:= -1;
  FCustomPaths:= TSTringList.Create;
  FPathGlyph:= -1;
  FFont := TFont.Create;
  FFont.Name := 'Tahoma';
  FFont.Color := clBlack;
  FFont.Size := 8;
  FMainBmp := TBitmap.Create;
  FMainBmp.PixelFormat := pf24Bit;
  FBorderStyle:= bsNone;
  FColorStart:= $00614F00;
  FColorEnd:= clAqua;
  FColorBorder:= $00C8BA90;
  FBrightness:= 255;
  FContrast:= 0;
  FCenter:= False;
  FReverse:= False;
  FPathIdx:= -1;
  FAuto:= True;
  PopupMenu:= Nil;
  SpecialFolders:= [spDocuments, spDesktop];

  barImageList:= TImagelist.Create(Nil);
  barImageList.ShareImages:= True;
  barImageList.DrawingStyle:= dsTransparent;

  ADesktopFolder := GetSpecialFolderPath(CSIDL_PERSONAL);
  TmpHandle:= SysImageListHandle(ADesktopFolder, False);
  if TmpHandle <> 0 then barImageList.Handle:= TmpHandle;

  StrMenu := TStringList.Create;

  popMenu:= TPopUpMenu.Create(nil);
  popMenu.OnPopup:= popMenuPopup;
  popMenu.Images:= barImageList;

  if not (csDesigning in ComponentState) then begin
    FEditPath:= TEdit.Create(Nil);
    FEditPath.Parent:= Self;
    FEditPath.ParentFont:= False;
    FEditPath.AlignWithMargins:= True;
    FEditPath.AutoSize:= False;
    FEDitPath.Margins.Left:= 1;
    FEDitPath.Margins.Top:= 1;
    FEDitPath.Margins.Right:= 1;
    FEDitPath.Margins.Bottom:= 1;
    FEditPath.Align:= alClient;
    FEditPath.Ctl3D:= True;
    FEditPath.Visible:= False;
    FEditPath.OnKeyDown:= EditKeyDown;
    FEditPath.OnKeyPress:= EditKeyPress;
    FEditPath.OnExit:= EditExit;
  end;

  FNewFolder:= 'NewFolder';
  Path := ADesktopFolder;
  Invalidate;
end;

destructor TrkSmartPath.Destroy;
begin
  FEditPath.Free;
  popMenu.Free;
  StrMenu.Free;
  barImageList.Free;
  FCustomPaths.Free;
  FPaths.Free;
  FFont.Free;
  FMainBMP.Free;
  inherited Destroy;
end;

procedure TrkSmartPath.Resize;
begin
  FPathOff:= CalcPath;
  Invalidate;
end;

procedure TrkSmartPath.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWord = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TrkSmartPath.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if (Button = mbLeft) then begin
    FHotPt:= Point(X, Y);
    CalcState;
    if (FState = spArrowHot) then FPopUp:= True else FPopUp:= False;
    FMClick:= True;
    Invalidate;
    if FAuto then begin
      if (FPopUp) and (FPopPath <> '') then begin
        popMenu.Popup(FPopPos.X, FPopPos.Y);
        PostMessage(Handle, CM_REFRESH, 0, 0);
      end;
    end else begin
      if (FPopUp) and Assigned(FCustomPopup) and (FPopPath <> '') then begin
        FCustomPopup.Popup(FPopPos.X, FPopPos.Y);
        PostMessage(Handle, CM_REFRESH, 0, 0);
      end;
    end;
    if (FState = spNormal) then DoEdit(False);
  end;
  Inherited;
end;

procedure TrkSmartPath.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  if FPopUp then Exit;
  FMClick:= False;
  FPopUp:= False;
  Invalidate;
  Inherited;
end;

procedure TrkSmartPath.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  FHotPt:= Point(X, Y);
  Invalidate;
end;

procedure TrkSmartPath.CMMouseLeave(var Message: TMessage);
begin
  if FPopUp then Exit;
  FMClick:= False;
  FPopUp:= False;
  FHotPt:= Point(-1, -1);
  Invalidate;
end;

procedure Register;
begin
  RegisterComponents( 'SmartView', [ TrkSmartPath ] );
end;

end.

